from browserexport.browsers.common import (
    handle_path,
    Paths,
)


from browserexport.browsers.chromium import Chromium


class Arc(Chromium):
    """https://arc.net/"""

    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                # macOS only browser, so no linux/darwin distinction
                "darwin": "~/Library/Application Support/Arc/User Data/",
            },
            browser_name=cls.__name__,
        )



import os
import sys
import sqlite3
import shutil
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Optional, Type, Union

import click
from sqlite_backup import sqlite_backup

from .log import logger
from .common import PathIsh, expand_path, BrowserexportError
from .browsers.all import DEFAULT_BROWSERS
from .browsers.common import Browser


def _progress(status: str, remaining: int, total: int) -> None:
    logger.debug(f"Copied {total - remaining} of {total} database pages...")


def _sqlite_backup(src: PathIsh, dest: Optional[Path]) -> Optional[sqlite3.Connection]:
    logger.info(f"backing up {src} to {dest}")
    return sqlite_backup(
        src,
        dest,
        wal_checkpoint=True,
        copy_use_tempdir=True,
        sqlite_backup_kwargs={"progress": _progress},
    )


def _print_sqlite_db_to_stdout(pth: Path) -> None:
    force = "BROWSEREXPORT_FORCE" in os.environ
    # make sure the user is piping this to something else, otherwise dont print
    if click.get_text_stream("stdout").isatty() and not force:
        logger.error(
            "stdout is a TTY, not printing database to stdout. Pipe to something else (e.g. browserexport save ... > db.sqlite, browserexport save ... | gzip --best > db.sqlite.gz) or set BROWSEREXPORT_FORCE=1 to print to stdout"
        )
        return

    logger.debug("writing bytes to stdout...")
    with open(pth, "rb") as f:
        shutil.copyfileobj(f, sys.stdout.buffer)  # type: ignore[misc]
        sys.stdout.buffer.flush()


def _path_backup(
    src: PathIsh,
    dest: PathIsh,
    browser_name: Optional[str] = None,
    pattern: Optional[str] = None,
) -> Optional[Path]:
    """
    Backup from src to dest. If dest is '-', print to stdout

    Otherwise, return the path to the backup
    """
    srcp: Path = expand_path(src)
    if str(dest) == "-":
        # use temporary directory as its more windows-friendly
        with tempfile.TemporaryDirectory() as td:
            tfp = Path(tempfile.mktemp(suffix="-browser-stdin.sqlite", dir=td))
            _sqlite_backup(srcp, tfp)
            _print_sqlite_db_to_stdout(tfp)

        assert not tfp.exists(), f"expected {tfp} to be deleted, but it still exists!"
        return None
    else:
        destp: Path = _default_pattern(
            srcp, dest, browser_name=browser_name, pattern=pattern
        )
        _sqlite_backup(srcp, destp)
        return destp


def _default_pattern(
    src: Path,
    to: PathIsh,
    browser_name: Optional[str] = None,
    pattern: Optional[str] = None,
) -> Path:
    """
    can pass a pattern with a format replacement field (for the timestamp)
    if you'd rather use a different format

    by default, this appends sqlite if that's not already the suffix,
    adds the name of the browser and a timestamp
    """
    to_p: Path = expand_path(to)
    assert to_p.is_dir(), f"{to_p} is not a directory!"
    suffix = src.suffix or ".sqlite"
    if pattern is None:
        pattern = (browser_name or "browser") + "-{}" + suffix
    # create pattern to timestamp backup filename
    now: str = datetime.utcnow().strftime("%Y%m%d%H%M%S")
    return to_p / pattern.format(now)


def backup_history(
    browser: Union[str, Type[Browser]],
    to: PathIsh,
    *,
    profile: str = "*",
    pattern: Optional[str] = None,
) -> Optional[Path]:
    """
    browser:
        typically the name of a browser the user provided from the CLI
        For library usage, user can pass a custom subclassed browser type which implements
            the required functions to locate the database
    to:
        path/string to backup database to. If '-', print to stdout
    profile:
        for browsers which have multiple profiles/users, this lets you specify
        a glob to select a particular profile
    pattern:
        pattern for the resulting timestamped filename, should include an str.format replacement placeholder

    returns:
        path to the backup, or None if printing to stdout
    """

    chosen: Type[Browser]
    browser_name: str
    if isinstance(browser, str):
        for extr in DEFAULT_BROWSERS:
            if browser.lower() == extr.__name__.lower():
                chosen = extr
                browser_name = browser.lower()
                break
        else:
            raise BrowserexportError(f"Unknown browser: {browser}")
    else:
        chosen = browser
        browser_name = chosen.__name__.lower()
    src: Path = chosen.locate_database(profile)
    dest: Optional[Path] = _path_backup(
        src, to, browser_name=browser_name, pattern=pattern
    )
    if str(to) == "-":
        assert (
            dest is None
        ), f"expected dest to be None since we printed to stdout, got {dest}"
    return dest



from .common import (
    Paths,
    handle_path,
    windows_appdata_paths,
)

from .chrome import Chrome


class EdgeDev(Chrome):
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "win32": windows_appdata_paths(r"Microsoft\Edge Dev"),
            },
            browser_name=cls.__name__,
        )



from browserexport.model import test_make_metadata
from browserexport.parse import _detect_extensions


def test_detect_extensions() -> None:
    assert _detect_extensions("foo.json") == ".json"
    assert _detect_extensions("foo.json.gz") == ".json"
    assert _detect_extensions("foo.jsonl") == ".jsonl"
    assert _detect_extensions("foo.jsonl.gz") == ".jsonl"
    assert _detect_extensions("/something/else/foo.sqlite") == ".sqlite"
    assert _detect_extensions("/something/else/foo.jsonl.gz") == ".jsonl"
    assert _detect_extensions("/something/else/foo.jsonl.zstd") == ".jsonl"
    assert _detect_extensions("/something/else/foo.jsonl.xz") == ".jsonl"



import os
import sqlite3
import tempfile
import shutil
from pathlib import Path
from typing import Iterator, List, Any, Dict, TextIO, Optional, Type, BinaryIO

from kompress import is_compressed, CPath

from .common import PathIshOrConn, PathIsh, expand_path, BrowserexportError
from .model import Visit
from .log import logger

from .browsers.common import Browser
from .browsers.all import DEFAULT_BROWSERS


def _read_json_obj(path: TextIO) -> Iterator[Dict[str, Any]]:
    data: List[Dict[str, Any]]
    try:
        # speedup load using orjson if its installed
        import orjson  # type: ignore[import]

        data = orjson.loads(path.read())

    except ImportError:
        import json

        data = json.loads(path.read())

    yield from data


def _read_json_file(path: PathIsh) -> Iterator[Dict[str, Any]]:
    with expand_path(path).open() as fp:
        yield from _read_json_obj(fp)


def _read_json_lines(fp: TextIO) -> Iterator[Dict[str, Any]]:
    try:
        import orjson  # type: ignore[import]

        for line in fp:
            yield orjson.loads(line)

    except ImportError:
        import json

        for line in fp:
            yield json.loads(line)


JSON_FORMATS = [".json", ".jsonl"]


def _detect_extensions(path: PathIsh) -> str:
    basepath: str = os.path.basename(str(path))
    if is_compressed(basepath):
        basepath, _, _ = basepath.rpartition(".")

    _, primary_ext = os.path.splitext(basepath)
    return primary_ext


def _parse_known_formats(path: PathIsh) -> Iterator[Visit]:
    pth: Path = CPath(expand_path(path))  # type: ignore
    ext = _detect_extensions(path)
    if ext not in JSON_FORMATS:
        raise ValueError(f"Unknown filetype: {path} extension={ext}")
    if ext == ".json":
        yield from map(Visit.from_dict, _read_json_file(pth))
    else:
        assert ext == ".jsonl", f"Unknown extension {ext}"
        logger.debug("Reading as JSON lines")
        with pth.open("r") as fp:
            yield from map(Visit.from_dict, _read_json_lines(fp))


def _read_buf_as_sqlite_db(buf: BinaryIO) -> sqlite3.Connection:
    """
    Reads some binary file object as sqlite database

    Pass sys.stdin.buffer to read from stdin
    """

    dbout = sqlite3.connect(":memory:")

    with tempfile.TemporaryDirectory() as td:
        tfp = Path(tempfile.mktemp(suffix="-browser-buffer.sqlite", dir=td))
        with tfp.open("wb") as tfo:
            logger.debug(f"reading buffer into tempfile {tfp}...")
            shutil.copyfileobj(buf, tfo)

        # copy to an in-memory sqlite database
        logger.debug("copying tempfile to in-memory sqlite database...")
        dbin = sqlite3.connect(str(tfp))

        # backup database, copy to in-memory database so that
        # once the tempfile is deleted, we can still access the data
        dbin.backup(dbout)

        dbin.close()

    assert not tfp.exists(), f"tempfile {tfp} should be deleted, but still exists"
    return dbout


def read_visits(
    path: PathIshOrConn, *, additional_browsers: Optional[List[Type[Browser]]] = None
) -> Iterator[Visit]:
    """
    Takes one sqlite database as input and returns 'Visit's
    """
    browsers: List[Type[Browser]] = additional_browsers or []
    browsers += DEFAULT_BROWSERS
    logger.info(f"Reading visits from {path}...")

    if isinstance(path, (str, Path)) and _detect_extensions(path) in JSON_FORMATS:
        logger.debug("Detected merged JSON file, mapping to Visit directly")
        try:
            yield from _parse_known_formats(path)
            return
        except ValueError as e:
            logger.debug(e, exc_info=True)
            logger.warning(
                f"Failed to parse {path} as known format, trying browsers instead"
            )

    for br in browsers:
        if br.detect(path):
            logger.debug(f"Detected as {br.__name__}")
            yield from br.extract_visits(path)
            return
    else:
        raise BrowserexportError(f"{path} didn't match any known schema")



from .common import Paths, handle_path, windows_appdata_paths


from .firefox import Firefox


class Librewolf(Firefox):
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": (
                    "~/.librewolf",
                    "~/.var/app/io.gitlab.librewolf-community/.librewolf",
                ),
                "darwin": "~/Library/Application Support/LibreWolf/Profiles",
                "win32": windows_appdata_paths(r"librewolf\Profiles"),
            },
            browser_name=cls.__name__,
        )



from .common import Paths, handle_path, windows_appdata_paths


from .firefox import Firefox


class Floorp(Firefox):
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": "~/.floorp",
                "darwin": "~/Library/Application Support/Floorp/Profiles",
                "win32": windows_appdata_paths(r"Floorp/Profiles"),
            },
            browser_name=cls.__name__,
        )



from typing import Optional
from .common import (
    BrowserexportError,
    logger,
    Paths,
    handle_path,
    windows_appdata_paths,
    handle_glob,
    Path,
)


from .chrome import Chrome


# TODO: Opera GX paths here or a separate file?
class Opera(Chrome):
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": "~/.config/opera",
                "darwin": "~/Library/Application Support/com.operasoftware.Opera",
                "win32": windows_appdata_paths(r"Opera Software\Opera Stable"),
            },
            browser_name=cls.__name__,
        )

    @classmethod
    def locate_database(cls, profile: str = "*") -> Path:
        # on linux, this seems to be just flat in ~/.config/opera without a 'Default' folder
        # like the typical chromium browsers. While on windows, its in the 'Default' folder like Chrome
        # can try both and see which works
        dd = cls.data_directories()
        err: Optional[BrowserexportError] = None
        # the '/' here allows the user to specify a profile name to disambiguate
        # but also makes it so it checks the base path + a subdir
        #
        # e.g. it checks:
        # ~/.config/opera/*/History
        # then
        # ~/.config/opera/*History
        #
        # If the user provides a profile name, it checks:
        # ~/.config/opera/ProfileName*/History
        # then
        # ~/.config/opera/*History
        for pth in ("/History", "History"):
            try:
                return handle_glob(dd, profile + pth)
            except BrowserexportError as e:
                logger.debug(e, exc_info=True)
                if err is None:
                    err = e

        assert err is not None
        raise err



from .common import (
    handle_path,
    windows_appdata_paths,
    Paths,
)


from .chrome import Chrome


class Chromium(Chrome):
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": (
                    "~/.config/chromium/",
                    "~/.var/app/org.chromium.Chromium/config/chromium/",
                    "~/snap/chromium/common/chromium/",
                ),
                "darwin": "~/Library/Application Support/Chromium/",
                "win32": windows_appdata_paths(r"Chromium\User Data"),
            },
            browser_name=cls.__name__,
        )



from typing import Union
from pathlib import Path
from sqlite3 import Connection

PathIsh = Union[str, Path]
# a path or a connection to a database
PathIshOrConn = Union[PathIsh, Connection]


def expand_path(path: PathIsh) -> Path:
    if isinstance(path, str):
        path = Path(path)
    return path.expanduser().absolute()


# keep as RuntimeError for backwards compatibility
class BrowserexportError(RuntimeError):
    pass



import sys
import os
import logging
import shlex
import json as jsn
from contextlib import contextmanager
from typing import List, Optional, Sequence, Iterator

import click

from .browsers.all import DEFAULT_BROWSERS
from .common import BrowserexportError
from .log import logger

CONTEXT_SETTINGS = {
    "max_content_width": 110,
    "show_default": True,
    "help_option_names": ["-h", "--help"],
}


# target for python3 -m browserexport and console_script using click
@click.group(
    context_settings=CONTEXT_SETTINGS,
    epilog="For more info, see https://github.com/seanbreckenridge/browserexport",
)
@click.option("--debug", is_flag=True, default=False, help="Increase log verbosity")
def cli(debug: bool) -> None:
    """
    Backup and merge your browser history!
    """
    from .log import setup

    if debug:
        setup(logging.DEBUG)


browsers_have_save: List[str] = [
    b.__name__.lower() for b in DEFAULT_BROWSERS if b.has_save
]

# shared click options
print_json = click.option(
    "-j",
    "--json",
    is_flag=True,
    default=False,
    required=False,
    help="Print result to STDOUT as JSON",
)

stream_json = click.option(
    "-s",
    "--stream",
    is_flag=True,
    default=False,
    required=False,
    help="Stream JSON objects instead of printing a JSON list",
)


@contextmanager
def _wrap_browserexport_cli_errors() -> Iterator[None]:
    try:
        yield
    except BrowserexportError as e:
        logger.debug(e, exc_info=True)
        click.echo(
            f"{click.style('Error:', 'red')} {e}, run as 'browserexport --debug' for more info"
        )
        exit(1)


def _chunk_list(lst: List[str], n: int) -> Iterator[List[str]]:
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def _wrapped_browser_list() -> str:
    import textwrap

    list_browsers_per_line = int(os.environ.get("LIST_BROWSERS_PER_LINE", 6))

    # split into groups of 6, join each group with ' | '
    lines_fmted: List[str] = [
        " | ".join(chunk)
        for chunk in _chunk_list(browsers_have_save, list_browsers_per_line)
    ]

    # add [ and ] to first and last
    lines_fmted[0] = "[" + lines_fmted[0]
    lines_fmted[-1] = lines_fmted[-1] + "]"

    # add | to the end of each line (separator between choices), except the last
    for i in range(0, len(lines_fmted) - 1):
        lines_fmted[i] = lines_fmted[i] + " |"

    return textwrap.indent("\n" + "\n".join(lines_fmted), " " * 6)


LIST_BROWSERS = "LIST_BROWSERS" in os.environ


@cli.command(
    epilog=(
        "For a list of all browsers, run 'LIST_BROWSERS=1 browserexport save --help'"
        if not LIST_BROWSERS
        else None
    ),
)
@click.option(
    "-b",
    "--browser",
    type=click.Choice(browsers_have_save, case_sensitive=False),
    metavar="BROWSER" if not LIST_BROWSERS else _wrapped_browser_list(),
    required=False,
    help="Browser name to backup history for",
)
@click.option(
    "--pattern",
    required=False,
    help="Pattern for the resulting timestamped filename, should include an str.format replacement placeholder for the date [default: browser_name-{}.extension]",
)
@click.option(
    "-p",
    "--profile",
    type=str,
    default="*",
    help="Use to pick the correct profile to back up. If unspecified, will assume a single profile",
)
@click.option(
    "--path",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Specify a direct path to a database to back up",
)
@click.option(
    "-t",
    "--to",
    type=click.Path(file_okay=False, allow_dash=True),
    required=True,
    help="Directory to store backup to. Pass '-' to print database to STDOUT",
)
@click.pass_context
def save(
    ctx: click.Context,
    browser: Optional[str],
    profile: str,
    to: str,
    path: Optional[str],
    pattern: Optional[str],
) -> None:
    """
    Backs up a current browser database file
    """
    from .save import backup_history, _path_backup

    if to != "-" and not os.path.exists(to):
        raise click.BadParameter(
            f"Invalid value for '-t' / '--to': Directory '{to}' does not exist"
        )

    if path is not None:
        if pattern is not None:
            click.echo(
                f"Warning: --pattern is not supported while using --path, if you want to backup to a specific path, you can use sqlite_backup directly:\n\npython3 -m sqlite_backup --debug {shlex.quote(path)} {shlex.quote(os.path.join(to, 'filename.sqlite'))}",
                err=True,
            )
        _path_backup(path, to)
    elif browser is not None:
        with _wrap_browserexport_cli_errors():
            backup_history(browser, to, profile=profile, pattern=pattern)
    else:
        click.secho(
            "Error: must provide one of '--browser', or '--path'\n",
            err=True,
            fg="red",
        )
        click.echo(ctx.get_help())


def _handle_merge(dbs: List[str], *, json: bool, stream: bool) -> None:
    from .model import Visit
    from .common import expand_path
    from .merge import merge_visits
    from .parse import read_visits, _read_buf_as_sqlite_db

    visits: List[Iterator[Visit]] = []

    with _wrap_browserexport_cli_errors():
        for db in dbs:
            if db == "-":
                logger.debug("Reading stdin as sqlite database")
                visits.append(read_visits(_read_buf_as_sqlite_db(sys.stdin.buffer)))
                continue
            # this is a command substitution, write it to a temp database so we can query against it
            # e.g. `browserexport merge <(browserexport save -b chrome -t -)`
            if os.path.islink(db) and os.readlink(db).startswith("pipe:"):
                logger.debug(f"Reading from proc file {db} into sqlite database")
                with open(db, "rb") as fp:
                    visits.append(read_visits(_read_buf_as_sqlite_db(fp)))
                continue
            if not os.path.exists(db):
                raise click.BadParameter(
                    f"Invalid value for SQLITE_DB: File '{db}' does not exist"
                )
            else:
                visits.append(read_visits(expand_path(db)))

        ivis: Iterator[Visit] = merge_visits(visits)
        if json or stream:
            if stream:
                for v in ivis:
                    sys.stdout.write(jsn.dumps(v.serialize(), separators=(",", ":")))
                    sys.stdout.write("\n")
                sys.stdout.flush()
            else:
                click.echo(
                    jsn.dumps([v.serialize() for v in ivis], separators=(",", ":"))
                )
        else:
            from .demo import demo_visit

            vis: List[Visit] = list(ivis)
            demo_visit(vis)
            header = f"Use {click.style('vis', fg='green')} to access visit data"

            try:
                import IPython  # type: ignore[import]
            except ModuleNotFoundError:
                click.secho(
                    "You may want to 'python3 -m pip install IPython' for a better REPL experience",
                    fg="yellow",
                )

                import code

                code.interact(local=locals(), banner=header)
            else:
                IPython.embed(header=header)  # type: ignore[no-untyped-call]


@cli.command()
@click.argument(
    "SQLITE_DB", type=click.Path(dir_okay=False, allow_dash=True), required=True
)
@stream_json
@print_json
def inspect(sqlite_db: str, json: bool, stream: bool) -> None:
    """
    Extracts visits from a single sqlite database

    \b
    Provide a history database as the first argument
    Drops you into a REPL to access the data

    Pass '-' to read from STDIN
    """
    with _wrap_browserexport_cli_errors():
        _handle_merge([sqlite_db], json=json, stream=stream)


@cli.command()
@click.argument(
    "SQLITE_DB",
    type=click.Path(dir_okay=False, allow_dash=True),
    nargs=-1,
    required=True,
)
@stream_json
@print_json
def merge(sqlite_db: Sequence[str], json: bool, stream: bool) -> None:
    """
    Extracts visits from multiple sqlite databases

    \b
    Provide multiple sqlite databases as positional arguments, e.g.:
    browserexport merge ~/data/firefox/*.sqlite

    Drops you into a REPL to access the data

    Pass '-' to read from STDIN
    """
    with _wrap_browserexport_cli_errors():
        _handle_merge(list(sqlite_db), json=json, stream=stream)


if __name__ == "__main__":
    cli(prog_name="browserexport")



from datetime import datetime, timezone
from pathlib import Path
from typing import Iterator, Sequence
import sqlite3

import pytest

from browserexport.common import expand_path
from browserexport.parse import read_visits
from browserexport.merge import read_and_merge


def test_using_conn(firefox: Path) -> None:
    conn = sqlite3.connect(f"file:{str(expand_path(firefox))}?immutable=1", uri=True)
    try:
        visits = list(read_visits(conn))
    finally:
        conn.close()
    assert len(visits) == 4


def test_read_visits(firefox: Path) -> None:
    vis = list(read_visits(firefox))
    assert len(vis) == 4
    vis = list(read_visits(str(firefox)))  # can also accept strings
    assert len(vis) == 4


def test_serialize_visit(firefox: Path) -> None:
    v = next(read_visits(firefox))
    sr_v = v.serialize()
    assert isinstance(sr_v, dict)
    assert set(sr_v.keys()) == {
        "metadata",
        "url",
        "dt",
    }
    assert sr_v["dt"] == 1593250194.51375
    assert sr_v["metadata"] is None
    assert sr_v["url"] == "https://www.mozilla.org/privacy/firefox/"


def test_merge_db(firefox: Path) -> None:
    # two of the same, should remove duplicates
    # and be the same as read_visits
    dbs: Sequence[Path] = [firefox, firefox]
    unique_count = len(list(read_and_merge(dbs)))
    direct_read_count = len(list(read_visits(firefox)))
    assert unique_count == direct_read_count


def test_read_chrome(chrome: Path) -> None:
    vis = list(read_visits(chrome))
    assert len(vis) == 6
    assert vis[0].metadata is not None
    assert vis[0].metadata.title == "Development Server"
    expected = datetime(2021, 1, 17, 6, 16, 15, 902496, tzinfo=timezone.utc)
    assert vis[0].dt == expected
    has_dur = vis[-1]
    assert has_dur.metadata is not None
    assert has_dur.metadata.duration == 16


def test_read_waterfox(waterfox: Path) -> None:
    vis = list(read_visits(waterfox))
    assert len(vis) == 2
    assert vis[0].url == "http://youtube.com/"
    expected = datetime(2021, 4, 16, 18, 36, 45, 691215, tzinfo=timezone.utc)
    assert vis[0].dt == expected


def test_read_chomium(chromium: Path) -> None:
    vis = list(read_visits(chromium))
    assert len(vis) == 1
    assert vis[0].url == "https://www.youtube.com/watch?v=dQw4w9WgXcQ"
    expected = datetime(2021, 4, 17, 19, 22, 2, 251822, tzinfo=timezone.utc)
    assert vis[0].dt == expected


def test_read_brave(brave: Path) -> None:
    vis = list(read_visits(brave))
    assert len(vis) == 2
    assert vis[0].metadata is not None
    assert vis[0].metadata.title == "User Terms of Service | Basic Attention Token"
    expected = datetime(2021, 4, 17, 19, 30, 32, 885828, tzinfo=timezone.utc)
    assert vis[0].dt == expected


def test_read_palemoon(palemoon: Path) -> None:
    vis = list(read_visits(palemoon))
    assert len(vis) == 8
    assert vis[0].metadata is not None
    assert vis[0].metadata.title == "Pale Moon - Congratulations"
    expected = datetime(2021, 4, 16, 18, 23, 23, 264033, tzinfo=timezone.utc)
    assert vis[0].dt == expected


def test_read_safari(safari: Path) -> None:
    vis = list(read_visits(safari))
    assert len(vis) == 3
    v = vis[-1]
    assert v.metadata is not None
    assert (
        v.metadata.title
        == "album amalgam - https://github.com/seanbreckenridge/albums - Google Sheets"
    )
    expected = datetime(2021, 4, 18, 1, 3, 45, 293084, tzinfo=timezone.utc)
    assert v.dt == expected


def test_read_vivaldi(vivaldi: Path) -> None:
    vis = list(read_visits(vivaldi))
    assert len(vis) == 3
    v = vis[-1]
    assert v.metadata is not None
    assert (
        v.metadata.title
        == "GitHub - seanbreckenridge/browserexport: backup and parse browser history databases"
    )
    expected = datetime(2021, 4, 19, 2, 26, 8, 29825, tzinfo=timezone.utc)
    assert v.dt == expected


def test_read_firefox_mobile_current(firefox_mobile: Path) -> None:
    vis = list(read_visits(firefox_mobile))
    assert len(vis) == 5
    v = vis[-1]
    assert v.url == "https://en.m.wikipedia.org/wiki/Vannevar_Bush"
    assert v.dt == datetime(2022, 2, 12, 8, 57, 19, 398000, tzinfo=timezone.utc)
    meta = v.metadata
    assert meta is not None
    assert meta.title == "Vannevar Bush - Wikipedia"
    assert meta.preview_image is not None


def test_read_firefox_mobile_legacy(firefox_mobile_legacy: Path) -> None:
    vis = list(read_visits(firefox_mobile_legacy))
    assert len(vis) == 3
    [v0, v1, v2] = vis
    assert v0.url == "https://github.com/"
    assert v0.dt == datetime(2020, 4, 11, 0, 54, 49, 44814, tzinfo=timezone.utc)
    assert v1.url == "https://xi.zulipchat.com/#"
    assert v1.dt == datetime(2020, 5, 28, 16, 18, 37, 6035, tzinfo=timezone.utc)
    v1meta = v1.metadata
    assert v1meta is not None, v1
    assert v1meta.title == "home - xi-editor and related projects - Zulip"
    assert v2.url == "https://github.com/"
    assert v2.dt == datetime(2020, 6, 27, 17, 34, 1, 963730, tzinfo=timezone.utc)


def test_merge_different(chrome: Path, waterfox: Path) -> None:
    chrome_vis = list(read_visits(chrome))
    waterfox_vis = list(read_visits(waterfox))
    merged_vis = list(read_and_merge([chrome, waterfox]))
    assert chrome_vis + waterfox_vis == merged_vis


def is_gz_file(filepath: Path) -> bool:
    with open(filepath, "rb") as test_f:
        # magic number for gzip files
        return test_f.read(2) == b"\x1f\x8b"


def test_read_json_dump(json_dump: Path) -> None:
    json_vis = list(read_visits(json_dump))
    assert not is_gz_file(json_dump)
    assert len(json_vis) == 1
    v = json_vis[0]
    assert v.url == "https://github.com/junegunn/fzf"
    assert v.dt == datetime(2020, 9, 15, 1, 29, 23, 720000, tzinfo=timezone.utc)
    assert v.metadata is not None
    assert v.metadata.preview_image == "https://github.com/favicon.ico"


def test_read_jsonl(jsonl_dump: Path) -> None:
    assert jsonl_dump.name.endswith(".jsonl")
    assert not is_gz_file(jsonl_dump)
    json_vis = list(read_visits(jsonl_dump))
    assert len(json_vis) == 3
    assert json_vis[0].url == "https://github.com/junegunn/fzf"
    assert json_vis[0].dt == datetime(
        2020, 9, 15, 1, 29, 23, 720000, tzinfo=timezone.utc
    )
    assert json_vis[1].url == "https://github.com/junegunn/fzf#installation"

    assert json_vis[1].metadata is not None
    assert json_vis[2].metadata is None


def test_read_json_gz(json_gz_dump: Path) -> None:
    assert json_gz_dump.name.endswith(".json.gz")
    assert is_gz_file(json_gz_dump)
    json_vis = list(read_visits(json_gz_dump))
    assert len(json_vis) == 1
    assert json_vis[0].url == "https://github.com/junegunn/fzf"


def test_read_jsonl_gz(jsonl_gz_dump: Path) -> None:
    assert jsonl_gz_dump.name.endswith(".jsonl.gz")
    assert is_gz_file(jsonl_gz_dump)
    json_vis = list(read_visits(jsonl_gz_dump))
    assert len(json_vis) == 2
    assert json_vis[0].url == "https://github.com/junegunn/fzf"
    assert json_vis[1].url == "https://github.com/junegunn/fzf#installation"


def test_reading_xz_file(xz_file: Path) -> None:
    assert xz_file.name.endswith(".json.xz")
    json_vis = list(read_visits(xz_file))
    assert len(json_vis) == 1
    assert json_vis[0].url == "https://github.com/junegunn/fzf"


def test_mixed_read(json_dump: Path, firefox: Path) -> None:
    jvis = list(read_visits(json_dump))
    fvisits = list(read_visits(firefox))
    assert len(jvis + fvisits) == 5
    assert len(list(read_and_merge([json_dump, firefox]))) == 5


databases_dir: Path = Path(__file__).parent / "databases"


def _database(name: str) -> Path:
    p = databases_dir / f"{name}.sqlite"
    assert p.exists()
    return p


# fixtures to make writing tests a bit easier
# though, maintaining them like this is quite annoying...


@pytest.fixture()
def json_dump() -> Iterator[Path]:
    p = databases_dir / "merged_dump.json"
    assert p.exists()
    yield p


@pytest.fixture()
def jsonl_dump() -> Iterator[Path]:
    p = databases_dir / "merged_dump.jsonl"
    assert p.exists()
    yield p


@pytest.fixture()
def json_gz_dump() -> Iterator[Path]:
    p = databases_dir / "merged_dump.json.gz"
    assert p.exists()
    yield p


@pytest.fixture()
def xz_file() -> Iterator[Path]:
    p = databases_dir / "merged_dump.json.xz"
    assert p.exists()
    yield p


@pytest.fixture()
def jsonl_gz_dump() -> Iterator[Path]:
    p = databases_dir / "merged_dump.jsonl.gz"
    assert p.exists()
    yield p


@pytest.fixture()
def firefox() -> Iterator[Path]:
    yield _database("firefox")


@pytest.fixture()
def chromium() -> Iterator[Path]:
    yield _database("chromium")


@pytest.fixture()
def chrome() -> Iterator[Path]:
    yield _database("chrome")


@pytest.fixture()
def brave() -> Iterator[Path]:
    yield _database("brave")


@pytest.fixture()
def palemoon() -> Iterator[Path]:
    yield _database("palemoon")


@pytest.fixture()
def waterfox() -> Iterator[Path]:
    yield _database("waterfox")


@pytest.fixture()
def safari() -> Iterator[Path]:
    yield _database("safari")


@pytest.fixture()
def vivaldi() -> Iterator[Path]:
    yield _database("vivaldi")


@pytest.fixture()
def firefox_mobile() -> Iterator[Path]:
    yield _database("firefox_mobile")


@pytest.fixture()
def firefox_mobile_legacy() -> Iterator[Path]:
    yield _database("firefox_mobile_legacy")



from .common import Paths, handle_path, windows_appdata_paths


from .chrome import Chrome


class Brave(Chrome):
    @classmethod
    def data_directories(cls) -> Paths:
        return handle_path(
            {
                "linux": (
                    "~/.config/BraveSoftware/Brave-Browser/",
                    "~/.var/app/com.brave.Browser/config/BraveSoftware/Brave-Browser/",
                ),
                "darwin": "~/Library/Application Support/BraveSoftware/Brave-Browser/",
                "win32": windows_appdata_paths(
                    r"BraveSoftware\Brave-Browser\User Data"
                ),
            },
            browser_name=cls.__name__,
        )



"""
Merges multiple history sqlite databases into one
"""

from datetime import datetime
from typing import Iterator, Iterable, Sequence, Set, Tuple, List

from .log import logger
from .model import Visit
from .common import PathIsh, expand_path
from .parse import read_visits


def read_and_merge(paths: Sequence[PathIsh]) -> Iterator[Visit]:
    """
    Receives any amount of Path-like databases as input,
    reads Visits from each of those databases,
    and merges them together (removing duplicates)
    """
    pths = [expand_path(p) for p in paths]
    hst: List[Iterator[Visit]] = list(map(read_visits, pths))
    yield from merge_visits(hst)


def merge_visits(sources: Sequence[Iterable[Visit]]) -> Iterator[Visit]:
    """
    Removes duplicate Visit items from multiple sources
    """
    logger.debug(f"merging information from {len(sources)} source(s)...")
    # use combination of URL, visit date and visit type to uniquely identify visits
    emitted: Set[Tuple[str, datetime]] = set()
    duplicates = 0
    for src in sources:
        for vs in src:
            key = (vs.url, vs.dt)
            if key in emitted:
                # logger.debug(f"skipping {key} => {vs}")
                duplicates += 1
                continue
            yield vs
            emitted.add(key)
    logger.debug("Summary: removed {} duplicates...".format(duplicates))
    logger.info("Summary: returning {} visit entries...".format(len(emitted)))



import os
import sys
import sqlite3

from itertools import chain
from pathlib import Path
from functools import lru_cache
from datetime import datetime, timezone
from typing import (
    Generator,
    List,
    Iterator,
    Optional,
    Dict,
    Union,
    Sequence,
)
from dataclasses import dataclass

import click

from ..log import logger
from ..model import Visit
from ..model import Metadata  # noqa: F401
from ..common import PathIsh, PathIshOrConn, expand_path, BrowserexportError
from ..sqlite import execute_query


@dataclass
class Schema:
    cols: List[str]
    where: str
    order_by: Optional[str] = None

    @property
    def query(self) -> str:
        qr = f"SELECT {', '.join(self.cols)} {self.where}"
        if self.order_by is not None:
            qr += f" ORDER BY {self.order_by}"
        return qr


Detector = str
Paths = Sequence[Path]


@dataclass
class Browser:
    schema: Schema  # used to create the query to extract visit from database
    detector: Detector  # semi-unique name of table, or a query to run on database to detect this type
    has_save: bool = True  # if this browser works with the save command

    @classmethod
    def detect(cls, path: PathIshOrConn) -> bool:
        """
        Run the detector against the given path/connection to detect if the current Browser matches the schema
        """
        # if the user provided something that had spaces - is a query
        if " " in cls.detector.strip():
            detector_query = cls.detector
        else:
            detector_query = f"SELECT * FROM {cls.detector}"
        logger.debug(f"{cls.__name__}: Running detector query '{detector_query}'")
        try:
            list(execute_query(path, detector_query))
            return True
        except sqlite3.OperationalError as sql_err:
            logger.debug(str(sql_err))
            return False

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        """
        Given a path or a sqlite3 connection, extract visits
        """
        raise NotImplementedError

    @classmethod
    def data_directories(cls) -> Paths:
        """
        The local data directories for this browser
        """
        raise NotImplementedError

    @classmethod
    def locate_database(cls, profile: str) -> Path:
        """
        Locate this database on the users' computer so it can be backed up
        """
        raise NotImplementedError


def from_datetime_microseconds(ts: int) -> datetime:
    return datetime.fromtimestamp(ts / 1_000_000, tz=timezone.utc)


errmsg = """Expected to match a single database, but found:
{}

You can use the --profile argument to select one of the profiles/match a particular file"""


def handle_glob(bases: Sequence[Path], stem: str, recursive: bool = False) -> Path:
    glob_itrs: List[Generator[Path, None, None]]
    if recursive:  # bleh, split like this to make mypy happy
        glob_itrs = [base.rglob(stem) for base in bases]
    else:
        glob_itrs = [base.glob(stem) for base in bases]
    dbs: List[Path] = list(chain(*glob_itrs))
    recur_desc = "recursive" if recursive else "non recursive"
    logger.debug(f"Glob {bases} with {stem} ({recur_desc}) matched {dbs}")
    if len(dbs) > 1:
        human_readable_db_paths: str = "\n".join([str(db) for db in dbs])
        raise BrowserexportError(errmsg.format(human_readable_db_paths))
    elif len(dbs) == 1:
        # found the match!
        return dbs[0]
    else:
        # if we weren't trying to search recursively, try a recursive search as a fallback
        if not recursive:
            return handle_glob(bases, stem, recursive=True)
        else:
            import shlex

            raise BrowserexportError(
                "Could not find database, using bases: '{bases}' and profile '{stem}'".format(
                    bases=", ".join(f'"{shlex.quote(str(base))}"' for base in bases),
                    stem=stem,
                )
            )


PROCFILE = Path("/proc/version")


@lru_cache(maxsize=1)
def determine_operating_system() -> str:
    # if this is being run from WSL, return win32
    # this should also detect cygwin as well
    if PROCFILE.exists():
        proc_version = PROCFILE.read_text()
        if "microsoft" in proc_version.lower():
            return "win32"
    return sys.platform


PathMapEntry = Union[PathIsh, Sequence[PathIsh]]


WINDOWS_BASE_ENVVARS = ("%LOCALAPPDATA%", "%APPDATA%")


# hmm, should this accept multiple paths?, callee could always
# just combine multiple calls to this function into one tuple
# and then pass that to handle_path
def windows_appdata_paths(path: str) -> Sequence[PathIsh]:
    """
    Given a path, return the path with the APPDATA/LOCALAPPDATA environment variables
    expanded.
    """
    # normpath converts the path to use the correct path separator for windows in case
    # the user provided a path with forward slashes
    return tuple(
        os.path.normpath(os.path.expandvars(os.path.join(envvar, path)))
        for envvar in WINDOWS_BASE_ENVVARS
    )


def handle_path(
    pathmap: Dict[str, PathMapEntry],
    browser_name: str,
    *,
    key: Optional[str] = None,
    default_behaviour: str = "linux",
) -> Paths:
    """
    Handles the repetitive task of having to resolve/expand a path
    which describes the location of the data directory on each
    operating system
    """
    loc: Sequence[PathIsh]
    # if the user didn't provide a key, assume this is a 'sys.platform' map - using
    # darwin/linux to specify the location
    if key is None:
        key = determine_operating_system()
    # use the key provided, or the first item (dicts after python3.7 are ordered)
    # in the pathmap if that doesn't exist
    maybeloc: Optional[PathMapEntry] = pathmap.get(key)
    if maybeloc is None:
        click.echo(
            f"""Not sure where {browser_name} history is installed on {key}
Defaulting to {default_behaviour} behaviour...

If you're using a browser/platform this currently doesn't support, please make an issue
at https://github.com/seanbreckenridge/browserexport/issues/new with information.
In the meantime, you can point this directly at a history database using the --path flag""",
            err=True,
        )
        maybeloc = pathmap[list(pathmap.keys())[0]]
    assert maybeloc is not None  # convince mypy
    if isinstance(maybeloc, (Path, str)):
        loc = [maybeloc]
    else:
        loc = maybeloc
    return tuple(expand_path(p) for p in loc)
